"""
iogp.data.carve: File carving / extraction APIs.

*coin regexes adapted from https://github.com/Concinnity-Risks/RansomCoinPublic

Author: Vlad Topan (vtopan/gmail)
"""
import mmap
import re
import struct


PAT = {
    'Bitcoin Private Key': '5[HJK][1-9A-Za-z][A-HJ-NP-Za-km-z]{48}',
    'Bitcoin Address': '[13][a-km-zA-HJ-NP-Z1-9]{25,34}',
    'Bitcoin Cash Address': '(bitcoincash:)?[qp][a-z0-9]{41}|(BITCOINCASH:)?[QP][A-Z0-9]{41}',
    'Ethereum Address': '0x[a-fA-F0-9]{40}',
    'Litecoin Address': '[LM3][a-km-zA-HJ-NP-Z1-9]{26,33}',
}
RX = {
    'file_magic': (rb'(7z)\xBC\xAF\x27\x1C|(PK)\x01\x02|(BM)|(MZ).{1,200}PE\0\0|\xFF\xD8\xFF\xE0..(JFIF)|\x7F(ELF)|(GIF)8[79]|\x89(PNG)\r\n\x1A\n|\xD0\xCF\x11\xE0(\xA1\xB1)\x1A\xE1|%(PDF)-1\.\d+\x0A', re.S),
    'text_file': (rb'[\x20-\x7E\x09\x0A\x0D]{512,}',),
    'jpg_end': (rb'\xFF\xD9',),
    'jpg_soi_end': (rb'\xFF[^\x00]',),
    'doc_end_1': ('\0'.join('DocumentSummaryInformation').encode(),),
    'doc_end_2': (rb'\0MSWordDoc\0\x10\0{3}Word.Document.8',),
    'pdf_end': (rb'[\r\n]%%EOF[\r\n]*',),
}
for k in RX:
    RX[k] = re.compile(*RX[k])

MAGIC_MAP = {
    # archive
    b'7z': '7z',
    b'PK': 'zip',
    # executable
    b'ELF': 'elf',
    b'MZ': 'exe',
    # image
    b'BM': 'bmp',
    b'GIF': 'gif',
    b'JFIF': 'jpg',
    b'PNG': 'png',
    # document
    b'\xA1\xB1': 'doc',
    b'PDF': 'pdf',
}

ZIP_POTENTIAL_CDS = set()


def extract_elf(data, offset):
    """
    Extract an ELF binary.
    """
    is64 = data[offset + 4] == 2
    spat, base, size = ('<QQIHHHHH', 0x20, 30) if is64 else ('<IIIHHHHH', 0x1C, 22)
    e_phoff, e_shoff, _, _, e_phentsize, e_phnum, e_shentsize, e_shnum  = struct.unpack(spat, data[offset + base:offset + base + size])
    szmax = 0
    for i in range(e_shnum):
        offs = e_shoff + e_shentsize * i
        offs, size = struct.unpack('<II', data[offs + 16:offs + 24])
        if size:
            szmax = max(szmax, offs + size)
    for i in range(e_phnum):
        offs = e_phoff + e_phentsize * i
        offs, _, _, size = struct.unpack('<IIII', data[offs + 4:offs + 20])
        if size:
            szmax = max(szmax, offs + size)
    return szmax, 'elf'


def extract_exe(data, offset):
    """
    Extract an MZ/PE image.
    """
    if offset + 0x400 > len(data):
        return None, None
    e_lfanew = struct.unpack('<i', data[offset + 0x3C:offset + 0x40])[0]
    peoffs = offset + e_lfanew
    if e_lfanew < 0 or peoffs + 0x1000 > len(data) or data[peoffs:peoffs + 4] != b'PE\0\0':
        return None, None
    seccnt = struct.unpack('<H', data[peoffs + 6:peoffs + 8])[0]
    secoffs = peoffs + 0xF8
    endoffs = -1
    for i in range(seccnt):
        offs = secoffs + i * 40 + 16
        fsize, faddr = struct.unpack('<II', data[offs:offs + 8])
        if fsize:
            endoffs = max(endoffs, faddr + fsize)
    return endoffs, 'exe'


def extract_zip(data, offset):
    """
    Extract a zip archive.
    """
    if offset in ZIP_POTENTIAL_CDS:
        return None, None
    # regex finds central directory entries, must validate each one - can mark checked ones though
    crt_offs = cd_start = offset
    delta = 28
    while 1:
        ZIP_POTENTIAL_CDS.add(crt_offs)
        base = crt_offs + delta
        fmt = '<HHHHHII'
        fmt_size = struct.calcsize(fmt)
        fn_len, ext_len, cmt_len, _, _, _, lfh_offs = struct.unpack(fmt, data[base:base + fmt_size])
        crt_offs += delta + fmt_size + fn_len + ext_len + cmt_len
        next_magic = data[crt_offs:crt_offs + 4]
        if next_magic[:2] != b'PK':
            return None, None
        if next_magic[2:4] != b'\x01\x02':
            if next_magic[2:4] == b'\x06\x06':
                # EOCDR64
                fmt = '<IQHHIIQQQQ'
                _, size, _, _, _, _, _, _, _, rel_offs = struct.unpack(fmt, data[crt_offs:crt_offs + struct.calcsize(fmt)])
                crt_offs += 12 + size
                next_magic = data[crt_offs:crt_offs + 4]
                if next_magic[:2] != b'PK':
                    return None, None
            if next_magic[2:4] == b'\x06\x07':
                # EOCDL64
                crt_offs += 20
                next_magic = data[crt_offs:crt_offs + 4]
                if next_magic[:2] != b'PK':
                    return None, None
            if next_magic[2:4] == b'\x05\x06':
                # EOCD
                fmt = '<IHHHHIIH'
                _, _, _, _, _, _, rel_offs, cmt_len = struct.unpack(fmt, data[crt_offs:crt_offs + struct.calcsize(fmt)])
                crt_offs += 22 + cmt_len
                start = cd_start - rel_offs
                if start < 0 or data[start:start + 4] != b'PK\x03\x04':
                    return None, None
                return (crt_offs - start, 'zip', start)
            else:
                return None, None


def extract_7z(data, offset):
    """
    Extract a 7z archive.
    """
    magic, ver_maj, ver_min, crc, next_offs, next_size, next_crc = struct.unpack('<6sBBIQQI', data[offset:offset + 32])
    if ver_maj != 0 or ver_min > 10:
        return None, None
    size = 32 + next_offs + next_size
    if offset + size > len(data):
        return None, None
    return size, '7z'


def extract_jpg(data, offset):
    """
    Extract a JPG image.
    """
    if offset == 0 or offset + 0x100 > len(data):
        return None, None
    end = offset + 2
    while end < len(data):
        hdr = data[end:end + 4]
        if len(hdr) == 2:
            hdr += b'\0\0'
        marker, size = struct.unpack(">HH", hdr)
        if marker == 0xffd9:
            break
        elif 0xffd0 <= marker <= 0xffda:
            pos = RX['jpg_soi_end'].search(data, pos=end + 4).start()
            end = pos
        else:
            end += size + 2
        if end - offset > 64 * 1024 * 1024:
            # this is probably a damaged JPG or false positive
            break
    return end - offset, 'jpg'


def extract_bmp(data, offset):
    """
    Extract a bitmap image.
    """
    magic, size, zero, offs = struct.unpack('<2sIII', data[offset:offset + 14])
    if zero != 0 or offset + offs > len(data) or offset + size > len(data):
        return None, None
    return size, 'bmp'


def extract_png(data, offset):
    """
    Extract a PNG image.
    """
    crt = offset + 8
    while crt + 8 < len(data):
        if crt - offset > 64 * 1024 * 1024:
            # this is probably a damaged PNG or false positive
            break
        size, ctype = struct.unpack('>I4s', data[crt:crt + 8])
        crt += size + 12
        if ctype == b'IEND':
            break
    return crt - offset, 'png'


def extract_doc(data, offset):
    """
    Extract a MS Word document.
    """
    if data[0x1A:0x1F] != b'\x03\0\xFE\xFF\x09':
        # todo: other docfiles
        return None, None
    if m := RX['doc_end_1'].search(data, pos=offset + 8, endpos=offset + 32 * 1024 * 1024):
        end = m.end() + 0x200
    elif m := RX['doc_end_2'].search(data, pos=offset + 8, endpos=offset + 32 * 1024 * 1024):
        end = m.end() + 0x100
    else:
        end = offset + 4 * 1024 * 1024
    return end - offset, 'doc'


def extract_pdf(data, offset):
    """
    Extract a PDF document.
    """
    if m := RX['pdf_end'].search(data, pos=offset + 8, endpos=offset + 32 * 1024 * 1024):
        end = m.end()
    else:
        end = offset + 4 * 1024 * 1024
    return end - offset, 'pdf'


def extract_embedded_files(data, whitelist=None, to_eof=False):
    """
    Extract embedded files from the given `bytes` buffer (or `mmap` object), returns

    :param whitelist: List of extensions to extract (default: all known).
    :return: Yields tuples (offset, size, extension); size is None if unknown.
    """
    global ZIP_POTENTIAL_CDS
    ZIP_POTENTIAL_CDS = set()
    for m in RX['file_magic'].finditer(data, pos=1):
        magic = [x for x in m.groups() if x][0]
        if whitelist and MAGIC_MAP[magic] not in whitelist:
            continue
        ext = MAGIC_MAP[magic]
        offset = m.start()
        extractor = globals().get('extract_' + ext)
        if extractor and not to_eof:
            res = extractor(data, offset)
            size, ext = res[:2]
            if len(res) == 3:
                offset = res[-1]
            if size == 0:
                size = len(data) - offset
        else:
            size = len(data) - offset
        if ext is not None:
            yield (offset, size, ext)


if __name__ == '__main__':
    import glob
    import os
    import sys

    whitelist = None if len(sys.argv) < 3 else [x.strip().lower() for x in sys.argv[2].split(',')]
    for f in glob.glob(sys.argv[1]):
        print(f'[*] Carving {f}...')
        fh = open(f, 'rb')
        data = mmap.mmap(f.fileno(), 0)
        for offs, size, ext in extract_embedded_files(data, whitelist):
            fn = f'{os.path.basename(f)}.emb@{offs:X}.{ext}'
            print(f'[-]   - found {ext} @ 0x{offs:X}[{size}] - saving as {fn}...')
            open(fn, 'wb').write(data[offs:offs + size])



"""
iogp.data.fileid: File identification APIs.

Author: Vlad Topan (vtopan/gmail)
"""

import re
import struct


HEADER_SIZE = 8192

FT_UNKNOWN = 'unknown'
FT_EMPTY = 'empty'

FT_REGEX = {
    # archive
    'svnz':         rb'^7z\xBC\xAF\x27\x1C',    # should be 7z, but Python regex named groups cannot contain digits
    'ace':          rb'^.{3}\0{3}.\*\*ACE\*\*\x14\x14',
    'alz':          rb'^ALZ\x01',
    'arj':          rb'^\x60\xEA',
    'bz2':          rb'^BZh',
    'cab_ishield':  rb'^ISc\x28',
    'cab_ms':       rb'^MSCF',
    'compressed':   rb'^(SZDD|KWAJ)\x88\xF0\x27(\x33|\xD1)',
    'crx':          rb'^Cr24',
    'deb':          rb'^!<arch>\.',
    'gz':           rb'^\x1F\x8B\x08',
    'lh_arc':       rb'^\x1A[\x02-\x04\x08\x09]',
    'lz':           rb'^LZIP',
    'lz4':          rb'^\x04"M\x18',
    'lzh':          rb'^..-lh',
    'lzs':          rb'^EDILZSS',
    'rar':          rb'^Rar!\x1A\x07',
    'rpm':          rb'^\xED\xAB\xEE\xDB',
    'tar':          rb'^.{256}\0ustar|^[\x20-\x7E].{64}\0{31}\0{4}([0-7 ]{6}[0-7 \0]\0){3}.{368}\0',
    'tar_z':        rb'^\x1F\x9d\x90',
    'xz':           rb'^\xFD7zXZ\0',
    'z':            rb'^\x13\x5De\x8C:\x01\x02\x00',
    'zlib':         rb'^\x78[\x01\x20\x5E\x7D\x9C\xBB\xDA\xF9]',
    'zoo':          rb'^ZOO ',
    # audiovideo
    'asf':          rb'^\x30\x26\xB2\x75\x8E\x66\xCF\x11',
    'flac':         rb'^fLaC',
    'mid':          rb'^MThd',
    'mkv':          rb'^\x1A\x45\xDF\xA3',
    'mp3':          rb'^\xFF[\xF2\xF3\xFB]|^ID3',
    'mpg':          rb'^\0\0\x01[\xBA\xB3]',
    'ogg':          rb'^OggS',
    'tgp':          rb'^ftyp3g',    # should be 3gp, but Python regex named groups cannot contain digits
    # disk
    'dmg':          rb'x\x01s\x0Dbb`',
    'iso':          rb'^CD001',
    'nes':          rb'^NES\x1A',
    'vmdk':         rb'^KDM',
    # doc
    'chm':          rb'^ITSF\x03\0{3}`\0{3}',
    'djvu':         rb'^AT&TFORM.{4}DJV',
    'doc':          rb'^\xD0\xCF\x11\xE0.{508}\xEC\xA5\xC1\0',
    'dwg':          rb'^AC(10|2\.)',
    'eml':          rb'^Received:',
    'html':         rb'(?i:^<(html|!doctype html))',
    'msg':          rb'^\xD0\xCF\x11\xE0.{508}' + 'Root Ent'.encode('utf16'),
    'one':          rb'\xE4\x52\x5C\x7B\x8C\xD8\xA7\x4D\xAE\xB1\x53\x78\xD0\x29\x96\xD3',
    'pdf':          rb'^%PDF-1',
    'ppt':          rb'^\xD0\xCF\x11\xE0.{508}(\x0F\0\xE8\x03|\xA0\x46\x1D\xF0|\0\x6E\x1E\xF0)',
    'ps':           rb'^\x04?%!PS',
    'rtf':          rb'^\{\\rtf',
    'swf':          rb'^[CF]WS',
    'wks':          rb'^\0{2}\x02\0\x04',
    'wordperfect':  rb'^\xFFWPC',
    'wri':          rb'^[12]\xBE\0{3}\xAB\0{2}',
    'xls':          rb'^\xD0\xCF\x11\xE0.{508}\x09\x08\x10\0\0\x06\x05\0',
    'xls_20':       rb'^\x09\0\x04\0\x02\0',
    'xml':          rb'^(?i:<\?xml)',
    # executable
    'class':        rb'^\xCA\xFE\xBA\xBE',
    'dex':          rb'^dex\x0A',
    'elf':          rb'^\x7FELF',
    'fox':          rb'^\xFB{2}.\x02',
    'lib_basic':    rb'^\xF0\x0D\0\0',
    'luac':         rb'^\x1BLua',
    'macho32':      rb'^\xCE\xFA\xED\xFE',
    'macho64':      rb'^\xCF\xFA\xED\xFE',
    'obj_coff':     rb'^L\x01',
    'pyc':          rb'^([\x6D\xB3\xD1]\xF2|\x03\xF3|[\x3B\x4F\x6C\x9E\xEE]\x0C|[\x16\x33\x42\x55]\x0D)\x0D\x0A',
    'wasm':         rb'^\0asm',
    # image
    'bmp':          rb'^BM',
    'bmp_ddb':      rb'^\x02\0{3}',
    'cdr':          rb'^WLe\0',
    'exr':          rb'^v/1\x01',
    'gif':          rb'^GIF8[79]',
    'ico':          rb'^\0\0(\x01\0|\x02\0)[\x01-\x04]\0',
    'jpg':          rb'^\xFF\xD8\xFF',
    'pcx':          rb'^\x0A[\x00-\x05]',
    'png':          rb'^\x89PNG\x0D\x0A\x1A\x0A',
    'psd':          rb'^8BPS',
    'svg':          rb'^<svg\s',
    'tif':          rb'^(II\x2A\0|MM\0\x2A)',
    'wmf':          rb'^\xD7\xCD\xC6\x9A',
    'wmf_3X':       rb'^\x01\0\x09\0',
    # installer files
    'ins':          rb'^(\xB8\xC9\x0C\0|\xFF{2}\x0C\0\x14\0\x34\x12)',
    'iss':          rb'^\[InstallSHIELD',
    'pkg':          rb'^\x4A\xA3',
    'stirling_exe': rb'^\x2A\xAB\x79\xD8\0\x01',
    # crypto
    'openssh_pub_key':  rb'ssh-\w+ [\w/+=]{32,}',
    'openssh_prv_key':  rb'-----BEGIN OPENSSH PRIVATE KEY-----',
    'pkcs8_prv_key':  rb'-----BEGIN PRIVATE KEY-----',
    'pkcs8_pub_key':  rb'-----BEGIN PUBLIC KEY-----',
    'pkcs1_prv_key':  rb'-----BEGIN [RD]SA PRIVATE KEY-----',
    'pkcs1_pub_key':  rb'-----BEGIN [RD]SA PUBLIC KEY-----',
    'ssh_rfc_prv_key':  rb'----[- ]BEGIN SSH2 PRIVATE KEY[- ]----',
    'ssh_rfc_pub_key':  rb'----[- ]BEGIN SSH2 PUBLIC KEY[- ]----',
    # misc
    'der':          rb'^0\x82',
    'evtx':         rb'^ElfFile',
    'hlp':          rb'^\x3F_\x03\0',
    'i64':          rb'^IDA2',
    'kdb':          rb'^7H\x03\x02\0{4}X509KEY',
    'lnk':          rb'^L\0{3}\x01\x14\x02\0{5}\xc0\0{6}F',
    'otf':          rb'^OTTO',
    'pcap':         rb'^\xA1\xB2\xC3\xD4|\xD4\xC3\xB2\xA1',
    'pcapng':       rb'^\x0A\x0D\x0D\x0A',
    'pdb':          rb'^Microsoft C/C\+\+ (program database|MSF) [\d\.]+\r\n\x1A',
    'pdb_ida':      rb'^IDA1',
    'pfb':          rb'^\x80\x01',
    'pgp':          rb'^(\x85\x01\x0C\x03|\x8C[\x04\x0C\x0D]\x04]|-----BEGIN PGP M)',
    'r1cs':         rb'^r1cs',
    'sqlite':       rb'^SQLite format',
    'ttc':          rb'^ttcf',
    'ttf':          rb'^\0\x01\0{3}.{3}\0[\x03\x04]\0.(OS/2|LTSH|cmap|DSIG|FFTM|GDEF|Feat|GPOS|ASCP|COLR|BASE|EBDT).{4}([\0\x01].{3}){2}[/A-Za-z0-9 ]{4}',
    # script
    'reg4':         rb'^REGEDIT4',
    'reg5':         rb'^\xFF\xFE' + 'Windows Registry Editor Version 5'.encode('utf16'),
    'script_posix': rb'^#!/[\w/ ]+',
    # keep at the end
    'text_utf8':    rb'^\xEF\xBB\xBF',
    'text_ascii':   rb'^[\x09\x0A\x0D\x20-\x7E]+\x1A?$',
    }
FULL_RX = re.compile(b'|'.join(b'(?P<%s>%s)' % (k.encode('utf8'), v) for k, v in FT_REGEX.items()), re.S)
FT_SUBTYPE_REGEX = {
    'svg':          rb'^(?i:<\?xml [^>]+\?>\s*(<\!(DOCTYPE|--)[^>]*>\s*)*<svg\s)',
    'vcproj':       rb'^.{0,50}<VisualStudioProject',
}
SUBTYPE_RX = re.compile(b'|'.join(b'(?P<%s>%s)' % (k.encode('utf8'), v) for k, v in FT_SUBTYPE_REGEX.items()), re.S)

FT_CATEGORY = {
    'audiovideo': ('3gp', 'asf', 'avi', 'flac', 'mid', 'mkv', 'mp3', 'mpg', 'ogg', 'wav',),
    'archive': ('7z', 'alz', 'arj', 'bz2', 'cab_ms', 'cab_ishield','compressed', 'crx', 'deb', 'dmg', 'egg', 'gz', 'jar',
            'lh_arc', 'lz', 'lz4', 'lzh', 'lzs', 'rar', 'rpm','tar', 'tar_z', 'xz', 'z', 'zip', 'zlib', 'zoo',),
    'disk': ('dmg', 'iso', 'nes', 'vmdk',),
    'document': ('chm', 'djvu', 'doc', 'dwg', 'eml', 'html', 'odt', 'one', 'pdf', 'ppt', 'ps', 'rtf', 'swf', 'text_ascii',
            'text_utf8', 'wks', 'wordperfect', 'wri', 'xls', 'xls_20', 'xml',),
    'executable': ('class', 'coff', 'dex', 'elf', 'fox', 'lib_basic', 'luac', 'macho32', 'macho64', 'mz', 'ne', 'obj_coff',
            'pe', 'pe+', 'wasm',),
    'image': ('ani', 'bmp', 'bmp_ddb', 'cdr', 'gif', 'ico', 'jpg', 'pcx', 'png', 'psd', 'tif', 'webp', 'wmf', 'wmf_3X',),
    'installers': ('ins', 'inst_exe', 'iss', 'pkg', 'stirling_exe',),
    'misc': ('der', 'fon', 'hlp', 'i64', 'kdb', 'lnk', 'otf', 'pcap', 'pcapng', 'pdb', 'pdb_ida', 'pfb', 'pgp', 'riff',
            'sqlite', 'ttc', 'ttf',),
    'script': ('reg4', 'reg5',),
    FT_UNKNOWN: (FT_UNKNOWN, FT_EMPTY, 'docfile_ms',),
    }
FT_CAT_MAP = {vv:k for k, v in FT_CATEGORY.items() for vv in v}

FT_EXT_MAP = {
    'elf': '',
    'macho32': 'so',
    'macho64': 'so',
    'mz': 'exe',
    'ne': 'exe',
    'pe': 'exe',
    'pe+': 'exe',
    'reg4': 'reg',
    'reg5': 'reg',
    'script_awk': 'awk',
    'script_node': 'js',
    'script_posix': '',
    'script_python': 'py',
    'script_perl': 'pl',
    'script_sh': 'sh',
    'sqlite': 'db',
    'stirling_exe': 'ex_',
    'tar_z': 'tar.z',
    'text_ascii': 'txt',
    'text_utf8': 'txt',
    'wordperfect': 'wpd',
    'thumbsdb': 'db',
    }

FT_2_MIME = {
    '7z': 'application/x-7z-compressed',
    'aac': 'audio/aac',
    'avi': 'video/x-msvideo',
    'azw': 'application/vnd.amazon.ebook',
    'bin': 'application/octet-stream',
    'bz2': 'application/x-bzip2',
    'css': 'text/css',
    'csv': 'text/csv',
    'doc': 'application/msword',
    'docx': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
    'epub': 'application/epub+zip',
    'gif': 'image/gif',
    'gz': 'application/gzip',
    'html': 'text/html',
    'ico': 'image/vnd.microsoft.icon',
    'ics': 'text/calendar',
    'jar': 'application/java-archive',
    'jpg': 'image/jpeg',
    'js': 'text/javascript',
    'json': 'application/json',
    'mp3': 'audio/mpeg',
    'mp4': 'video/mp4',
    'odp': 'application/vnd.oasis.opendocument.presentation',
    'ods': 'application/vnd.oasis.opendocument.spreadsheet',
    'odt': 'application/vnd.oasis.opendocument.text',
    'ogg': 'audio/ogg',
    'pdf': 'application/pdf',
    'php': 'application/x-httpd-php',
    'png': 'image/png',
    'ppt': 'application/vnd.ms-powerpoint',
    'pptx': 'application/vnd.openxmlformats-officedocument.presentationml.presentation',
    'rar': 'application/vnd.rar',
    'rtf': 'application/rtf',
    'sh': 'application/x-sh',
    'svg': 'image/svg+xml',
    'tar': 'application/x-tar',
    'tiff': 'image/tiff',
    'ttf': 'font/ttf',
    'txt': 'text/plain',
    'wav': 'audio/wav',
    'weba': 'audio/webm',
    'webm': 'video/webm',
    'webp': 'image/webp',
    'xhtml': 'application/xhtml+xml',
    'xls': 'application/vnd.ms-excel',
    'xlsx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
    'xml': 'application/xml',
    'zip': 'application/zip',
}
MIME_2_FT = {v: k for k, v in FT_2_MIME.items()}


def get_file_type(source):
    """
    Identify the type of a file (or `bytes` buffer).

    :param source: File handle, filename or bytes buffer.
    """
    if hasattr(source, 'read'):
        header = source.read(HEADER_SIZE)
    elif type(source) == str:
        header = open(source, 'rb').read(HEADER_SIZE)
    else:
        header = source[:HEADER_SIZE]
    if len(header) == 0:
        return FT_EMPTY
    if header[:2] == b'MZ':
        if len(header) >= 0x100:
            elfanew = struct.unpack('<L', header[0x3C:0x40])[0]
            if elfanew + 8 < len(header):
                if header[elfanew:elfanew + 4] == b'PE\0\0':
                    machine = header[elfanew + 4:elfanew + 6]
                    if machine == b'\x64\x86':
                        return 'pe+'
                    elif machine == b'\x4c\x01':
                        return 'pe'
                    else:
                        return 'coff'
            ne_offs = struct.unpack('<H', header[0x3C:0x3E])[0]
            if ne_offs + 0x30 < len(header):
                if header[ne_offs:ne_offs + 2] == b'NE':
                    if b'FONTRES' in header[ne_offs + 0x80:ne_offs + 0x140]:
                        # todo: parse NE header
                        return 'fon'
                    return 'ne'
        return 'mz'
    elif header[:4] == b'RIFF':
        mark = header[8:12]
        if mark == b'WAVE':
            return 'wav'
        elif mark == b'AVI ':
            return 'avi'
        elif mark == b'ACON':
            return 'ani'
        elif mark == b'WEBP':
            return 'webp'
        else:
            return 'riff'
    elif header[:2] == b'PK':
        if b'META-INF/MANIFEST' in header and b'.class' in header:
            return 'jar'
        elif b'mimetypeapplication/epub+zip' in header[:0x40]:
            return 'epub'
        elif b'mimetypeapplication/vnd.oasis.opendocument.tex' in header[:0x50]:
            return 'odt'
        elif b'EGG-INFO/' in header[:0x40]:
            return 'egg'
        elif b'[Content_Types].xml' in header[:0x50]:
            # M$ Office zipped XML format
            if b'\x01xl/_rels/' in header:
                return 'xlsx'
            if b'\x01word/_rels/' in header:
                return 'docx'
            return 'office-zip'
        return 'zip'
    m = FULL_RX.search(header)
    if m:
        for k in FT_REGEX:
            if m.group(k):
                match = m.group()
                if k == 'tgp':
                    k = '3gp'
                elif k == 'svnz':
                    k = '7z'
                elif k in ('xml',):
                    if sm := SUBTYPE_RX.search(header):
                        for sk in FT_SUBTYPE_REGEX:
                            if sm.group(sk):
                                k = sk
                                break
                elif k == 'script_posix':
                    for e in (b'perl', b'python', b'make', b'awk', b'wish', b'node'):
                        if re.search(rb'[ /]%s($|\W)' % e, match):
                            e = e.decode('utf8')
                            return 'script_' + {'wish': 'tcl'}.get(e, e)
                    if re.search(rb'/[tck]?sh(\W|$)', match):
                        return 'script_sh'
                return k
    if header[:4] == b'\xD0\xCF\x11\xE0':
        return 'docfile_ms'
    if len(header) == 0x200 and header[0x1FE:0x200] == b'\x55\xAA':
        return 'bootsector_x86'
    return FT_UNKNOWN


def get_ft_category(filetype):
    """
    Returns the category of the given filetype.
    """
    if filetype.startswith('script_'):
        return 'script'
    return FT_CAT_MAP.get(filetype, FT_UNKNOWN)


def get_ft_ext(filetype):
    """
    Returns the most common file extension for the given filetype (or FT_UNKNOWN).
    """
    ext = FT_EXT_MAP.get(filetype, filetype)
    if '_' in ext and not ext.endswith('_'):
        ext = ext.split('_', 1)[0]
    return ext
